/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 *
 *  Copyright © 2006 Pierre Habouzit
 */

/*
 * Copyright notice from original mutt:
 * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>

#include <lib-lib/mem.h>
#include <lib-lib/str.h>
#include <lib-lib/ascii.h>
#include <lib-lib/macros.h>
#include <lib-lib/buffer.h>
#include <lib-lib/date.h>

#include <lib-crypt/crypt.h>

#include "recvattach.h"
#include "mx.h"
#include "url.h"

#include "lib/debug.h"

#include "mime.h"

/* Reads an arbitrarily long header field, and looks ahead for continuation
 * lines.  ``line'' must point to a dynamically allocated string; it is
 * increased if more space is required to fit the whole line.
 */
ssize_t mutt_read_rfc822_line(FILE *f, char **line, ssize_t *n)
{
    ssize_t pos = 0;

    for (;;) {
        char *p = *line;

        /* end of file or end of headers */
        if (!fgets(p + pos, *n - pos, f) || (ISSPACE(*p) && pos == 0)) {
            *p = '\0';
            return 0;
        }

        pos += m_strlen(p + pos);
        if (p[pos - 1] == '\n') {
            int c;

            /* remove trailing spaces. safe: p[0] is not a space */
            do {
                p[--pos] = '\0';
            } while (ISSPACE(p[pos]));

            /* check to see if the next line is a continuation line */
            c = fgetc(f);
            if (c != ' ' && c != '\t') {
                /* next line is a separate header field or EOH */
                ungetc(c, f);
                return pos;
            }

            /* eat tabs and spaces from the beginning of the continuation line */
            do {
                c = fgetc(f);
            } while (c == ' ' || c == '\t');
            ungetc(c, f);

            /* string is still terminated because we removed at least one
               whitespace char above */
            p[pos++] = ' ';
        }

        if (*n < pos + STRING) {
            /* grow the buffer */
            *n += STRING;
            p_realloc(line, *n);
        }
    }
}

/* TODO: Make that a string list somehow */
LIST *mutt_parse_references(char *s, int in_reply_to)
{
    LIST *lst = NULL;
    int n = 0;
    char *o = NULL;

    /* some mail clients add other garbage besides message-ids, so do a quick
     * check to make sure this looks like a valid message-id
     * some idiotic clients also break their message-ids between lines, deal
     * with that too (give up if it's more than two lines, though)
     */

    for (s = strtok(s, " \t;"); s; s = strtok(NULL, " \t;")) {
        char *new = NULL;

        if (*s == '<') {
            n = m_strlen(s);
            if (s[n - 1] != '>') {
                o = s;
                continue;
            }

            new = m_strdup(s);
        } else if (o) {
            ssize_t m = m_strlen(s);

            if (s[m - 1] != '>') {
                o = NULL;
            } else {
                new = p_new(char, n + m + 1);
                strcpy(new, o);
                strcpy(new + n, s);
            }
        }

        /* make sure that this really does look like a message-id.
         * it should have exactly one @, and if we're looking at
         * an in-reply-to header, make sure that the part before
         * the @ has more than eight characters or it's probably
         * an email address
         */
        if (new) {
            char *at = strchr(new, '@');
            LIST *tmp;

            if (!at || strchr(at + 1, '@') || (in_reply_to && at - new <= 8)) {
                p_delete(&new);
                continue;
            }

            tmp = p_new(LIST, 1);
            tmp->data = new;
            tmp->next = lst;
            lst = tmp;
        }
    }

    return lst;
}

int mutt_check_encoding(const char *s)
{
#define COMPARE(tok, value)                             \
    if (!ascii_strncasecmp(tok, s, sizeof(tok) - 1)) {  \
        return value;                                   \
    }
    COMPARE("7bit", ENC7BIT);
    COMPARE("8bit", ENC8BIT);
    COMPARE("binary", ENCBINARY);
    COMPARE("quoted-printable", ENCQUOTEDPRINTABLE);
    COMPARE("base64", ENCBASE64);
    COMPARE("x-uuencode", ENCUUENCODED);
#undef COMPARE

    return ENCOTHER;
}

int mutt_check_mime_type(const char *s)
{
#define COMPARE(tok, value)                             \
    if (!ascii_strncasecmp(tok, s, sizeof(tok) - 1)) {  \
        return value;                                   \
    }
  COMPARE("text", TYPETEXT);
  COMPARE("multipart", TYPEMULTIPART);
  COMPARE("application", TYPEAPPLICATION);
  COMPARE("message", TYPEMESSAGE);
  COMPARE("image", TYPEIMAGE);
  COMPARE("audio", TYPEAUDIO);
  COMPARE("video", TYPEVIDEO);
  COMPARE("model", TYPEMODEL);
  COMPARE("*",  TYPEANY);
  COMPARE(".*", TYPEANY);
#undef COMPARE

  return TYPEOTHER;
}

static PARAMETER *parse_parameters(const char *s)
{
    PARAMETER *res = NULL;
    PARAMETER **list = &res;

    while (*s) {
        const char *p;
        PARAMETER *new;
        int i;

        s = skipspaces(s);
        if (*s == '=')             /* parameters are fucked up, go away */
            break;

        p = strpbrk(s, "=;");
        if (!p)
            break;

        if (*p == ';') {
            /* if we hit a ; now the parameter has no value, just skip it */
            s = p + 1;
            continue;
        }

        i = p - s;
        new = parameter_new();
        new->attribute = p_dupstr(s, i);

        while (--i >= 0 && ISSPACE(new->attribute[i])) {
            new->attribute[i] = '\0';
        }
        s = skipspaces(p + 1);                      /* skip over the = */

        if (*s == '"') {
            char buffer[LONG_STRING];
            int state_ascii = 1;

            s++;
            for (i = 0; *s && i < ssizeof(buffer) - 1; i++, s++) {
                if (!option(OPTSTRICTMIME)) {
                    /* As iso-2022-* has a characer of '"' with non-ascii state,
                     * ignore it. */
                    if (*s == 0x1b && i < ssizeof(buffer) - 2) {
                        state_ascii = s[1] == '(' && (s[2] == 'B' || s[2] == 'J');
                    }
                }
                if (state_ascii && *s == '"')
                    break;

                if (*s == '\\') {
                    buffer[i] = *++s;
                } else {
                    buffer[i] = *s;
                }
            }

            new->value = p_dupstr(buffer, i);
        } else {
            for (p = s; *p && *p != ' ' && *p != ';'; p++);
            new->value = p_dupstr(s, p - s);
        }

        *list = new;
        list = &new->next;

        s = strchr(s, ';');           /* Find the next parameter */
        if (!s)
            break;                    /* no more parameters */
    }

    rfc2231_decode_parameters(&res);
    return res;
}

void mutt_parse_content_type(char *s, BODY *ct)
{
    char *pc;
    char *subtype;

    p_delete(&ct->subtype);
    parameter_delete(&ct->parameter);

    /* First extract any existing parameters */
    if ((pc = strchr(s, ';')) != NULL) {
        *pc++ = '\0';
        ct->parameter = parse_parameters(vskipspaces(pc));

        /* Some pre-RFC1521 gateways still use the "name=filename" convention,
         * but if a filename has already been set in the content-disposition,
         * let that take precedence, and don't set it here */
        pc = mutt_get_parameter("name", ct->parameter);
        if (pc && !ct->filename)
            ct->filename = m_strdup(pc);
    }

    /* Now get the subtype */
    if ((subtype = strchr (s, '/'))) {
        *subtype++ = '\0';
        for (pc = subtype; *pc && !ISSPACE(*pc) && *pc != ';'; pc++);
        ct->subtype = p_dupstr(subtype, pc - subtype);
    }

    /* Finally, get the major type */
    ct->type = mutt_check_mime_type(s);

    if (ct->type == TYPEOTHER) {
        ct->xtype = m_strdup(s);
    }

    if (!ct->subtype) {
        /* Some older non-MIME mailers (i.e., mailtool, elm) have a content-type
         * field, so we can attempt to convert the type to BODY here.
         */
        switch (ct->type) {
            char buffer[SHORT_STRING];

          case TYPETEXT:
            ct->subtype = m_strdup("plain");
            break;

          case TYPEAUDIO:
            ct->subtype = m_strdup("basic");
            break;

          case TYPEMESSAGE:
            ct->subtype = m_strdup("rfc822");
            break;

          case TYPEOTHER:
            ct->type = TYPEAPPLICATION;
            snprintf(buffer, sizeof(buffer), "x-%s", s);
            ct->subtype = m_strdup(buffer);
            break;

          default:
            ct->subtype = m_strdup("x-unknown");
            break;
        }
    }

    /* Default character set for text types. */
    if (ct->type == TYPETEXT) {
        pc = mutt_get_parameter("charset", ct->parameter);
        if (!pc) {
            mutt_set_parameter("charset",
                               option(OPTSTRICTMIME) ? "us-ascii" :
                               mutt_get_first_charset(AssumedCharset),
                               &ct->parameter);
        }
    }
}

static void parse_content_disposition(char *s, BODY *ct)
{
    if (!ascii_strncasecmp(s, "inline", 6)) {
        ct->disposition = DISPINLINE;
    } else if (!ascii_strncasecmp(s, "form-data", 9)) {
        ct->disposition = DISPFORMDATA;
    } else {
        ct->disposition = DISPATTACH;
    }

    /* Check to see if a default filename was given */
    if ((s = strchr (s, ';'))) {
        PARAMETER *parms = parse_parameters(vskipspaces(s));

        if ((s = mutt_get_parameter("filename", parms)))
            m_strreplace(&ct->filename, s);
        if ((s = mutt_get_parameter ("name", parms)))
            ct->form_name = m_strdup(s);

        parameter_delete(&parms);
    }
}

/* args:
 *	fp	stream to read from
 *
 *	digest	1 if reading subparts of a multipart/digest, 0
 *		otherwise
 */
BODY *mutt_read_mime_header(FILE *fp, int digest)
{
    BODY *body = mutt_new_body ();
    char *line = p_new(char, LONG_STRING);
    ssize_t linelen = LONG_STRING;
    char *p;

    body->hdr_offset  = ftello(fp);
    body->encoding    = ENC7BIT;    /* default from RFC1521 */
    body->disposition = DISPINLINE;
    body->type        = digest ? TYPEMESSAGE : TYPETEXT;

    while (mutt_read_rfc822_line(fp, &line, &linelen)) {
        /* Find the value of the current header */
        if ((p = strchr(line, ':'))) {
            *p++ = '\0';
            p = vskipspaces(p);
            if (!*p)
                continue;
        } else {
            debug_print (1, ("bogus MIME header: %s\n", line));
            break;
        }

        if (!ascii_strncasecmp(line, "content-", 8)) {
            if (!ascii_strcasecmp("type", line + 8))
                mutt_parse_content_type (p, body);
            else if (!ascii_strcasecmp ("transfer-encoding", line + 8))
                body->encoding = mutt_check_encoding (p);
            else if (!ascii_strcasecmp ("disposition", line + 8))
                parse_content_disposition(p, body);
            else if (!ascii_strcasecmp ("description", line + 8)) {
                m_strreplace(&body->description, p);
                rfc2047_decode(&body->description);
            }
        }
    }

    body->offset = ftello(fp);       /* Mark the start of the real data */
    if (!body->subtype) {
        if (body->type == TYPETEXT)
            body->subtype = m_strdup("plain");
        if (body->type == TYPEMESSAGE)
            body->subtype = m_strdup("rfc822");
    }

    p_delete(&line);
    return (body);
}

void mutt_parse_part(FILE *fp, BODY *b)
{
    char *bound = 0;

    switch (b->type) {
      case TYPEMULTIPART:
        bound = mutt_get_parameter("boundary", b->parameter);
        fseeko(fp, b->offset, SEEK_SET);
        b->parts = mutt_parse_multipart(fp, bound, b->offset + b->length,
                           !ascii_strcasecmp("digest", b->subtype));
        break;

      case TYPEMESSAGE:
        if (b->subtype) {
            fseeko(fp, b->offset, SEEK_SET);

            if (mutt_is_message_type(b->type, b->subtype)) {
                b->parts = mutt_parse_messageRFC822(fp, b);
            } else
            if (!ascii_strcasecmp(b->subtype, "external-body") == 0) {
                b->parts = mutt_read_mime_header(fp, 0);
            } else {
                return;
            }
        }
        break;

      default:
        return;
    }

    /* try to recover from parsing error */
    if (!b->parts) {
        b->type = TYPETEXT;
        m_strreplace(&b->subtype, "plain");
    }
}

/* parse a MESSAGE/RFC822 body
 *
 * args:
 *	fp		stream to read from
 *
 *	parent		structure which contains info about the message/rfc822
 *			body part
 *
 * NOTE: this assumes that `parent->length' has been set!
 */
BODY *mutt_parse_messageRFC822(FILE * fp, BODY * parent)
{
    BODY *msg;

    parent->hdr = header_new();
    parent->hdr->offset = ftello(fp);
    parent->hdr->env    = mutt_read_rfc822_header(fp, parent->hdr, 0, 0);

    msg = parent->hdr->content;

    /* ignore the length given in the content-length since it could be wrong
       and we already have the info to calculate the correct length */
    /* if (msg->length == -1) */
    /* if body of this message is empty, we can end up with a negative length */
    msg->length = MAX(0, parent->length - (msg->offset - parent->offset));

    mutt_parse_part(fp, msg);

    return msg;
}

/* parse a multipart structure
 *
 * args:
 *	fp		stream to read from
 *
 *	bound   	body separator
 *
 *	end_off		length of the multipart body (used when the final
 *			boundary is missing to avoid reading too far)
 *
 *	digest		1 if reading a multipart/digest, 0 otherwise
 */

BODY *
mutt_parse_multipart(FILE *fp, const char *bound, off_t end_off, int digest)
{
    char buffer[LONG_STRING];
    BODY *head = NULL;
    BODY **last = &head;
    int blen = m_strlen(bound);
    int final = 0;                /* did we see the ending boundary? */

    if (!blen) {
        mutt_error _("multipart message has no boundary parameter!");
        return NULL;
    }

    while (ftello(fp) < end_off && fgets(buffer, sizeof(buffer), fp)) {
        int len, crlf, i;

        len  = m_strlen(buffer);
        crlf = len > 1 && buffer[len - 2] == '\r';

        if (buffer[0] == '-' && buffer[1] == '-'
        && !m_strncmp(buffer + 2, bound, blen))
        {
            if (*last) {
                BODY *b = *last;

                /* if the body is empty, we can end up with a -1 length */
                b->length = MAX(0, ftello(fp) - b->offset - len - 1 - crlf);
                if (b->parts && b->parts->length == 0) {
                    b->parts->length = ftello(fp) - b->parts->offset
                                     - len - 1 - crlf;
                }
            }

            /* Remove any trailing whitespace, up to the length of the boundary */
            for (i = len - 1; ISSPACE(buffer[i]) && i >= blen + 2; i--)
                buffer[i] = '\0';

            /* Check for the end boundary */
            final = buffer[blen + 3] == '-' && buffer[blen + 4] == '-';
            if (final)
                break;

            if (buffer[2 + blen] == '\0') {
                BODY *new = mutt_read_mime_header(fp, digest);

                /*
                 * Consistency checking - catch
                 * bad attachment end boundaries
                 */

                if (new->offset > end_off) {
                    mutt_free_body(&new);
                    break;
                }

                if (*last)
                    last = &(*last)->next;
                *last = new;
            }
        }
    }

    /* in case of missing end boundary, set the length to something reasonable */
    if (*last && (*last)->length == 0 && !final)
        (*last)->length = end_off - (*last)->offset;

    /* parse recursive MIME parts */
    {
        BODY *b;
        for (b = head; b; b = b->next)
            mutt_parse_part(fp, b);
    }

    return (head);
}

static const char *
uncomment_timezone(char *buf, size_t buflen, const char *tz)
{
    char *p;

    if (*tz != '(')
        return tz;                  /* no need to do anything */

    tz = vskipspaces(tz + 1);
    p = strpbrk(tz, " )");
    if (!p)
        return tz;

    m_strncpy(buf, buflen, tz, p - tz);
    return buf;
}

static struct tz_t {
    char tzname[5];
    unsigned char zhours;
    unsigned char zminutes;
    unsigned char zoccident;      /* west of UTC? */
} TimeZones[] = {
    {"aat", 1, 0, 1},             /* Atlantic Africa Time */
    {"adt", 4, 0, 0},             /* Arabia DST */
    {"ast", 3, 0, 0},             /* Arabia */
    /*{ "ast",   4,  0, 1 }, *//* Atlantic */
    {"bst", 1, 0, 0},             /* British DST */
    {"cat", 1, 0, 0},             /* Central Africa */
    {"cdt", 5, 0, 1},
    {"cest", 2, 0, 0},            /* Central Europe DST */
    {"cet", 1, 0, 0},             /* Central Europe */
    {"cst", 6, 0, 1},
    /*{ "cst",   8,  0, 0 }, *//* China */
    /*{ "cst",   9, 30, 0 }, *//* Australian Central Standard Time */
    {"eat", 3, 0, 0},             /* East Africa */
    {"edt", 4, 0, 1},
    {"eest", 3, 0, 0},            /* Eastern Europe DST */
    {"eet", 2, 0, 0},             /* Eastern Europe */
    {"egst", 0, 0, 0},            /* Eastern Greenland DST */
    {"egt", 1, 0, 1},             /* Eastern Greenland */
    {"est", 5, 0, 1},
    {"gmt", 0, 0, 0},
    {"gst", 4, 0, 0},             /* Presian Gulf */
    {"hkt", 8, 0, 0},             /* Hong Kong */
    {"ict", 7, 0, 0},             /* Indochina */
    {"idt", 3, 0, 0},             /* Israel DST */
    {"ist", 2, 0, 0},             /* Israel */
    /*{ "ist",   5, 30, 0 }, *//* India */
    {"jst", 9, 0, 0},             /* Japan */
    {"kst", 9, 0, 0},             /* Korea */
    {"mdt", 6, 0, 1},
    {"met", 1, 0, 0},             /* this is now officially CET */
    {"msd", 4, 0, 0},             /* Moscow DST */
    {"msk", 3, 0, 0},             /* Moscow */
    {"mst", 7, 0, 1},
    {"nzdt", 13, 0, 0},           /* New Zealand DST */
    {"nzst", 12, 0, 0},           /* New Zealand */
    {"pdt", 7, 0, 1},
    {"pst", 8, 0, 1},
    {"sat", 2, 0, 0},             /* South Africa */
    {"smt", 4, 0, 0},             /* Seychelles */
    {"sst", 11, 0, 1},            /* Samoa */
    /*{ "sst",   8,  0, 0 }, *//* Singapore */
    {"utc", 0, 0, 0},
    {"wat", 0, 0, 0},             /* West Africa */
    {"west", 1, 0, 0},            /* Western Europe DST */
    {"wet", 0, 0, 0},             /* Western Europe */
    {"wgst", 2, 0, 1},            /* Western Greenland DST */
    {"wgt", 3, 0, 1},             /* Western Greenland */
    {"wst", 8, 0, 0},             /* Western Australia */
};

/* parses a date string in RFC822 format:
 *
 * Date: [ weekday , ] day-of-month month year hour:minute:second timezone
 *
 * This routine assumes that `h' has been initialized to 0.  the `timezone'
 * field is optional, defaulting to +0000 if missing.
 */
time_t mutt_parse_date(const char *s, HEADER *h)
{
    int count = 0;
    char *p;
    struct tm tm;
    int tz_offset = 0;
    int zhours = 0, zminutes = 0, zoccident = 0;
    const char *ptz;
    char tzstr[SHORT_STRING];
    char scratch[SHORT_STRING];

    /* Don't modify our argument. Fixed-size buffer is ok here since
       the date format imposes a natural limit.  */

    m_strcpy(scratch, sizeof(scratch), s);

    /* kill the day of the week, if it exists. */
    p = strchr(scratch, ',');
    p = vskipspaces(p ? p + 1 : scratch);

    p_clear(&tm, 1);

    while ((p = strtok (p, " \t")) != NULL) {
        switch (count) {
          case 0:                    /* day of the month */
            if (!isdigit((unsigned char)*p))
                return -1;
            tm.tm_mday = atoi(p);
            if (tm.tm_mday > 31)
                return -1;
            break;

          case 1:                    /* month of the year */
            tm.tm_mon = mutt_check_month(p);
            if (tm.tm_mon < 0)
                return -1;
            break;

          case 2:                    /* year */
            tm.tm_year = atoi(p);
            if (tm.tm_year < 50)
                tm.tm_year += 100;
            else if (tm.tm_year >= 1900)
                tm.tm_year -= 1900;
            break;

          case 3:                    /* time of day */
            tm.tm_hour = strtol(p, &p, 10);
            if (*p++ != ':')
                return -1;
            tm.tm_min  = strtol(p, &p, 10);
            if (*p++ == ':') {
                tm.tm_sec = strtol(p, &p, 10);
            } else {
                tm.tm_sec = 0;
            }
            break;

          case 4:                    /* timezone */
            /* sometimes we see things like (MST) or (-0700) so attempt to
             * compensate by uncommenting the string if non-RFC822 compliant
             */
            ptz = uncomment_timezone(tzstr, sizeof(tzstr), p);

            if (*ptz == '+' || *ptz == '-') {
                if (isdigit((unsigned char)ptz[1])
                &&  isdigit((unsigned char)ptz[2])
                &&  isdigit((unsigned char)ptz[3])
                &&  isdigit((unsigned char)ptz[4]))
                {
                    zoccident = ptz[0] == '-';
                    zhours    = (ptz[1] - '0') * 10 + (ptz[2] - '0');
                    zminutes  = (ptz[3] - '0') * 10 + (ptz[4] - '0');
                }
            } else {
                struct tz_t *tz;

                /* This is safe to do: A pointer to a struct equals a pointer to its
                 * first element*/
                tz = bsearch(ptz, TimeZones, countof(TimeZones), sizeof(TimeZones[0]),
                             (int (*)(const void *, const void *))ascii_strcasecmp);

                if (tz) {
                    zhours = tz->zhours;
                    zminutes = tz->zminutes;
                    zoccident = tz->zoccident;
                }

                /* ad hoc support for the European MET (now officially CET) TZ */
                if (ascii_strcasecmp(p, "MET") == 0) {
                    if ((p = strtok (NULL, " \t")) && !ascii_strcasecmp(p, "DST")) {
                        zhours++;
                    }
                }
            }

            tz_offset = zhours * 3600 + zminutes * 60;
            if (!zoccident)
                tz_offset = -tz_offset;
            break;
        }
        count++;
        p = NULL;
    }

    if (count < 4) {  /* don't check for missing timezone */
        debug_print (1, ("error parsing date format, using received time\n"));
        return -1;
    }

    if (h) {
        h->zhours    = zhours;
        h->zminutes  = zminutes;
        h->zoccident = zoccident;
    }

    return mutt_mktime(&tm, 0) + tz_offset;
}

/*** XXX: MC READ MARK ***/

/* extract the first substring that looks like a message-id */
static char *extract_message_id(const char *s)
{
    const char *p;

    if ((s = strchr(s, '<')) == NULL || (p = strchr(s, '>')) == NULL)
        return NULL;
    return p_dupstr(s, (p - s) + 1);
}

void mutt_parse_mime_message (CONTEXT * ctx, HEADER * cur)
{
  MESSAGE *msg;
  int flags = 0;

  do {
    if (cur->content->type != TYPEMESSAGE
        && cur->content->type != TYPEMULTIPART)
      break;                     /* nothing to do */

    if (cur->content->parts)
      break;                     /* The message was parsed earlier. */

    if ((msg = mx_open_message (ctx, cur->msgno))) {
      mutt_parse_part (msg->fp, cur->content);

      cur->security = crypt_query (cur->content);

      mx_close_message (&msg);
    }
  } while (0);
  mutt_count_body_parts (cur, flags | M_PARTS_RECOUNT);
}

int mutt_parse_rfc822_line (ENVELOPE * e, HEADER * hdr, char *line, char *p,
                            short user_hdrs, short weed, short do_2047,
                            LIST ** lastp)
{
  int matched = 0;
  LIST *last = NULL;

  if (lastp)
    last = *lastp;

  switch (ascii_tolower (line[0])) {
  case 'a':
    if (ascii_strcasecmp (line + 1, "pparently-to") == 0) {
      e->to = rfc822_parse_adrlist (e->to, p);
      matched = 1;
    }
    else if (ascii_strcasecmp (line + 1, "pparently-from") == 0) {
      e->from = rfc822_parse_adrlist (e->from, p);
      matched = 1;
    }
    break;

  case 'b':
    if (ascii_strcasecmp (line + 1, "cc") == 0) {
      e->bcc = rfc822_parse_adrlist (e->bcc, p);
      matched = 1;
    }
    break;

  case 'c':
    if (ascii_strcasecmp (line + 1, "c") == 0) {
      e->cc = rfc822_parse_adrlist (e->cc, p);
      matched = 1;
    }
    else if (ascii_strncasecmp (line + 1, "ontent-", 7) == 0) {
      if (ascii_strcasecmp (line + 8, "type") == 0) {
        if (hdr)
          mutt_parse_content_type (p, hdr->content);
        matched = 1;
      }
      else if (ascii_strcasecmp (line + 8, "transfer-encoding") == 0) {
        if (hdr)
          hdr->content->encoding = mutt_check_encoding (p);
        matched = 1;
      }
      else if (ascii_strcasecmp (line + 8, "length") == 0) {
        if (hdr) {
          if ((hdr->content->length = atoi (p)) < 0)
            hdr->content->length = -1;
        }
        matched = 1;
      }
      else if (ascii_strcasecmp (line + 8, "description") == 0) {
        if (hdr) {
          m_strreplace(&hdr->content->description, p);
          rfc2047_decode (&hdr->content->description);
        }
        matched = 1;
      }
      else if (ascii_strcasecmp (line + 8, "disposition") == 0) {
        if (hdr)
          parse_content_disposition (p, hdr->content);
        matched = 1;
      }
    }
    break;

  case 'd':
    if (!ascii_strcasecmp ("ate", line + 1)) {
      m_strreplace(&e->date, p);
      if (hdr)
        hdr->date_sent = mutt_parse_date (p, hdr);
      matched = 1;
    }
    break;

  case 'e':
    if (!ascii_strcasecmp ("xpires", line + 1) &&
        hdr && mutt_parse_date (p, NULL) < time (NULL))
      hdr->expired = 1;
    break;

  case 'f':
    if (!ascii_strcasecmp ("rom", line + 1)) {
      e->from = rfc822_parse_adrlist (e->from, p);
      /* don't leave from info NULL if there's an invalid address (or
       * whatever) in From: field; mutt would just display it as empty
       * and mark mail/(esp.) news article as your own. aaargh! this
       * bothered me for _years_ */
      if (!e->from) {
        e->from = address_new ();
        e->from->personal = m_strdup(p);
      }
      matched = 1;
    }
#ifdef USE_NNTP
    else if (!m_strcasecmp(line + 1, "ollowup-to")) {
      if (!e->followup_to) {
        m_strrtrim(p);
        e->followup_to = m_strdup(skipspaces(p));
      }
      matched = 1;
    }
#endif
    break;

  case 'i':
    if (!ascii_strcasecmp (line + 1, "n-reply-to")) {
      mutt_free_list (&e->in_reply_to);
      e->in_reply_to = mutt_parse_references (p, 1);
      matched = 1;
    }
    break;

  case 'l':
    if (!ascii_strcasecmp (line + 1, "ines")) {
      if (hdr) {
        hdr->lines = atoi (p);

        /*
         * HACK - mutt has, for a very short time, produced negative
         * Lines header values.  Ignore them.
         */
        if (hdr->lines < 0)
          hdr->lines = 0;
      }

      matched = 1;
    }
    else if (!ascii_strcasecmp (line + 1, "ist-Post")) {
      /* RFC 2369.  FIXME: We should ignore whitespace, but don't. */
      if (strncmp (p, "NO", 2)) {
        char *beg, *end;

        for (beg = strchr (p, '<'); beg; beg = strchr (end, ',')) {
          ++beg;
          if (!(end = strchr (beg, '>')))
            break;

          /* Take the first mailto URL */
          if (url_check_scheme (beg) == U_MAILTO) {
            p_delete(&e->list_post);
            e->list_post = p_dupstr(beg, end - beg);
            break;
          }
        }
      }
      matched = 1;
    }
    break;

  case 'm':
    if (!ascii_strcasecmp (line + 1, "ime-version")) {
      if (hdr)
        hdr->mime = 1;
      matched = 1;
    }
    else if (!ascii_strcasecmp (line + 1, "essage-id")) {
      /* We add a new "Message-ID:" when building a message */
      p_delete(&e->message_id);
      e->message_id = extract_message_id (p);
      matched = 1;
    }
    else if (!ascii_strncasecmp (line + 1, "ail-", 4)) {
      if (!ascii_strcasecmp (line + 5, "reply-to")) {
        /* override the Reply-To: field */
        address_delete (&e->reply_to);
        e->reply_to = rfc822_parse_adrlist (e->reply_to, p);
        matched = 1;
      }
      else if (!ascii_strcasecmp (line + 5, "followup-to")) {
        e->mail_followup_to = rfc822_parse_adrlist (e->mail_followup_to, p);
        matched = 1;
      }
    }
    break;

#ifdef USE_NNTP
  case 'n':
    if (!m_strcasecmp(line + 1, "ewsgroups")) {
      p_delete(&e->newsgroups);
      m_strrtrim(p);
      e->newsgroups = m_strdup(skipspaces(p));
      matched = 1;
    }
    break;
#endif

  case 'o':
    /* field `Organization:' saves only for pager! */
    if (!m_strcasecmp(line + 1, "rganization")) {
      if (!e->organization && m_strcasecmp(p, "unknown"))
        e->organization = m_strdup(p);
    }
    break;

  case 'r':
    if (!ascii_strcasecmp (line + 1, "eferences")) {
      mutt_free_list (&e->references);
      e->references = mutt_parse_references (p, 0);
      matched = 1;
    }
    else if (!ascii_strcasecmp (line + 1, "eply-to")) {
      e->reply_to = rfc822_parse_adrlist (e->reply_to, p);
      matched = 1;
    }
    else if (!ascii_strcasecmp (line + 1, "eturn-path")) {
      e->return_path = rfc822_parse_adrlist (e->return_path, p);
      matched = 1;
    }
    else if (!ascii_strcasecmp (line + 1, "eceived")) {
      if (hdr && !hdr->received) {
        char *d = strchr (p, ';');

        if (d)
          hdr->received = mutt_parse_date (d + 1, NULL);
      }
    }
    break;

  case 's':
    if (!ascii_strcasecmp (line + 1, "ubject")) {
      if (!e->subject)
        e->subject = m_strdup(p);
      matched = 1;
    }
    else if (!ascii_strcasecmp (line + 1, "ender")) {
      e->sender = rfc822_parse_adrlist (e->sender, p);
      matched = 1;
    }
    else if (!ascii_strcasecmp (line + 1, "tatus")) {
      if (hdr) {
        while (*p) {
          switch (*p) {
          case 'r':
            hdr->replied = 1;
            break;
          case 'O':
            hdr->old = 1;
            break;
          case 'R':
            hdr->read = 1;
            break;
          }
          p++;
        }
      }
      matched = 1;
    }
    else if ((!ascii_strcasecmp ("upersedes", line + 1) ||
              !ascii_strcasecmp ("upercedes", line + 1)) && hdr)
      e->supersedes = m_strdup(p);
    break;

  case 't':
    if (ascii_strcasecmp (line + 1, "o") == 0) {
      e->to = rfc822_parse_adrlist (e->to, p);
      matched = 1;
    }
    break;

  case 'x':
    if (ascii_strcasecmp (line + 1, "-status") == 0) {
      if (hdr) {
        while (*p) {
          switch (*p) {
          case 'A':
            hdr->replied = 1;
            break;
          case 'D':
            hdr->deleted = 1;
            break;
          case 'F':
            hdr->flagged = 1;
            break;
          default:
            break;
          }
          p++;
        }
      }
      matched = 1;
    }
    else if (ascii_strcasecmp (line + 1, "-label") == 0) {
      e->x_label = m_strdup(p);
      matched = 1;
    }
#ifdef USE_NNTP
    else if (!m_strcasecmp(line + 1, "-comment-to")) {
      if (!e->x_comment_to)
        e->x_comment_to = m_strdup(p);
      matched = 1;
    }
    else if (!m_strcasecmp(line + 1, "ref")) {
      if (!e->xref)
        e->xref = m_strdup(p);
      matched = 1;
    }
#endif

  default:
    break;
  }

  /* Keep track of the user-defined headers */
  if (!matched && user_hdrs) {
    /* restore the original line */
    line[m_strlen(line)] = ':';

    if (weed && option (OPTWEED) && mutt_matches_ignore (line, Ignore)
        && !mutt_matches_ignore (line, UnIgnore))
      goto done;

    if (last) {
      last->next = mutt_new_list ();
      last = last->next;
    }
    else
      last = e->userhdrs = mutt_new_list ();
    last->data = m_strdup(line);
    if (do_2047)
      rfc2047_decode (&last->data);
  }

done:

  *lastp = last;
  return matched;
}


/* mutt_read_rfc822_header() -- parses a RFC822 header
 *
 * Args:
 *
 * f		stream to read from
 *
 * hdr		header structure of current message (optional).
 *
 * user_hdrs	If set, store user headers.  Used for recall-message and
 * 		postpone modes.
 *
 * weed		If this parameter is set and the user has activated the
 * 		$weed option, honor the header weed list for user headers.
 * 	        Used for recall-message.
 *
 * Returns:     newly allocated envelope structure.  You should free it by
 *              envelope_delete() when envelope stay unneeded.
 */
ENVELOPE *mutt_read_rfc822_header (FILE * f, HEADER * hdr, short user_hdrs,
                                   short weed)
{
  ENVELOPE *e = envelope_new();
  LIST *last = NULL;
  char *line = p_new(char, LONG_STRING);
  char *p;
  off_t loc;
  int matched;
  ssize_t linelen = LONG_STRING;
  char buf[LONG_STRING + 1];

  if (hdr) {
    if (hdr->content == NULL) {
      hdr->content = mutt_new_body ();

      /* set the defaults from RFC1521 */
      hdr->content->type = TYPETEXT;
      hdr->content->subtype = m_strdup("plain");
      hdr->content->encoding = ENC7BIT;
      hdr->content->length = -1;

      /* RFC 2183 says this is arbitrary */
      hdr->content->disposition = DISPINLINE;
    }
  }

  while ((loc = ftello (f)),
         mutt_read_rfc822_line (f, &line, &linelen))
  {
    matched = 0;

    if ((p = strpbrk (line, ": \t")) == NULL || *p != ':') {
      char return_path[LONG_STRING];
      time_t t;

      /* some bogus MTAs will quote the original "From " line */
      if (m_strncmp(">From ", line, 6) == 0)
        continue;               /* just ignore */
      else if (is_from (line, return_path, sizeof (return_path), &t)) {
        /* MH somtimes has the From_ line in the middle of the header! */
        if (hdr && !hdr->received)
          hdr->received = t - mutt_local_tz (t);
        continue;
      }

      fseeko (f, loc, 0);
      break;                    /* end of header */
    }

    *buf = '\0';

    if (mutt_match_spam_list (line, SpamList, buf, sizeof (buf))) {
      if (!rx_list_match (NoSpamList, line)) {

        /* if spam tag already exists, figure out how to amend it */
        if (e->spam && *buf) {
          /* If SpamSep defined, append with separator */
          if (SpamSep) {
            mutt_buffer_addstr (e->spam, SpamSep);
            mutt_buffer_addstr (e->spam, buf);
          }

          /* else overwrite */
          else {
            e->spam->dptr = e->spam->data;
            *e->spam->dptr = '\0';
            mutt_buffer_addstr (e->spam, buf);
          }
        }

        /* spam tag is new, and match expr is non-empty; copy */
        else if (!e->spam && *buf) {
          e->spam = mutt_buffer_from (NULL, buf);
        }

        /* match expr is empty; plug in null string if no existing tag */
        else if (!e->spam) {
          e->spam = mutt_buffer_from (NULL, "");
        }

        if (e->spam && e->spam->data)
          debug_print (5, ("spam = %s\n", e->spam->data));
      }
    }

    *p++ = 0;
    p = vskipspaces(p);
    if (!*p)
      continue;                 /* skip empty header fields */

    matched =
      mutt_parse_rfc822_line (e, hdr, line, p, user_hdrs, weed, 1, &last);

  }

  p_delete(&line);

  if (hdr) {
    hdr->content->hdr_offset = hdr->offset;
    hdr->content->offset = ftello (f);
    rfc2047_decode_envelope(e);
    /* check for missing or invalid date */
    if (hdr->date_sent <= 0) {
      debug_print (1, ("no date found, using received "
                       "time from msg separator\n"));
      hdr->date_sent = hdr->received;
    }
  }

  return (e);
}

address_t *mutt_parse_adrlist (address_t * p, const char *s)
{
  const char *q;

  /* check for a simple whitespace separated list of addresses */
  if ((q = strpbrk (s, "\"<>():;,\\")) == NULL) {
    char tmp[HUGE_STRING];
    char *r;

    m_strcpy(tmp, sizeof(tmp), s);
    r = tmp;
    while ((r = strtok (r, " \t")) != NULL) {
      p = rfc822_parse_adrlist (p, r);
      r = NULL;
    }
  }
  else
    p = rfc822_parse_adrlist (p, s);

  return p;
}


/* Compares mime types to the ok and except lists */
int count_body_parts_check(LIST **checklist, BODY *b, int dflt) {
  LIST *type;
  ATTACH_MATCH *a;

  /* If list is null, use default behavior. */
  if (! *checklist) {
    /*return dflt;*/
    return 0;
  }

  for (type = *checklist; type; type = type->next) {
    a = (ATTACH_MATCH *)type->data;
    debug_print(5, ("cbpc: %s %d/%s ?? %s/%s [%d]... ",
               dflt ? "[OK] " : "[EXCL] ",
               b->type, b->subtype, a->major, a->minor, a->major_int));
    if ((a->major_int == TYPEANY || a->major_int == b->type) &&
        !regexec(&a->minor_rx, b->subtype, 0, NULL, 0)) {
      debug_print(5, ("yes\n"));
      return 1;
    } else {
      debug_print(5, ("no\n"));
    }
  }
  return 0;
}

#define AT_COUNT(why) { shallcount = 1; }
#define AT_NOCOUNT(why) { shallcount = 0; }

int count_body_parts (BODY *body, int flags) {
  int count = 0;
  int shallcount, shallrecurse;
  BODY *bp;

  if (body == NULL)
    return 0;

  for (bp = body; bp != NULL; bp = bp->next) {
    /* Initial disposition is to count and not to recurse this part. */
    AT_COUNT("default");
    shallrecurse = 0;

    debug_print(5, ("bp: desc=\"%s\"; fn=\"%s\", type=\"%d/%s\"\n",
               bp->description ? bp->description : ("none"),
               bp->filename ? bp->filename :
               bp->d_filename ? bp->d_filename : "(none)",
               bp->type, bp->subtype ? bp->subtype : "*"));

    if (bp->type == TYPEMESSAGE) {
      shallrecurse = 1;

      /* If it's an external body pointer, don't recurse it. */
      if (!ascii_strcasecmp (bp->subtype, "external-body"))
        shallrecurse = 0;

      /* Don't count containers if they're top-level. */
      if (flags & M_PARTS_TOPLEVEL)
        AT_NOCOUNT("top-level message/*");
    } else if (bp->type == TYPEMULTIPART) {
      /* Always recurse multiparts, except multipart/alternative. */
      shallrecurse = 1;
      if (!m_strcasecmp(bp->subtype, "alternative"))
        shallrecurse = 0;

      /* Don't count containers if they're top-level. */
      if (flags & M_PARTS_TOPLEVEL)
        AT_NOCOUNT("top-level multipart");
    }

    if (bp->disposition == DISPINLINE &&
        bp->type != TYPEMULTIPART && bp->type != TYPEMESSAGE && bp == body)
      AT_NOCOUNT("ignore fundamental inlines");

    /* If this body isn't scheduled for enumeration already, don't bother
     * profiling it further. */

    if (shallcount) {
      /* Turn off shallcount if message type is not in ok list,
       * or if it is in except list. Check is done separately for
       * inlines vs. attachments.
       */

      if (bp->disposition == DISPATTACH) {
        if (!count_body_parts_check(&AttachAllow, bp, 1))
          AT_NOCOUNT("attach not allowed");
        if (count_body_parts_check(&AttachExclude, bp, 0))
          AT_NOCOUNT("attach excluded");
      } else {
        if (!count_body_parts_check(&InlineAllow, bp, 1))
          AT_NOCOUNT("inline not allowed");
        if (count_body_parts_check(&InlineExclude, bp, 0))
          AT_NOCOUNT("excluded");
      }
    }

    if (shallcount)
      count++;
    bp->attach_qualifies = shallcount ? 1 : 0;

    debug_print(5, ("cbp: %p shallcount = %d\n", bp, shallcount));

    if (shallrecurse) {
      debug_print(5, ("cbp: %p pre count = %d\n", bp, count));
      bp->attach_count = count_body_parts(bp->parts, flags & ~M_PARTS_TOPLEVEL);
      count += bp->attach_count;
      debug_print(5, ("cbp: %p post count = %d\n", bp, count));
    }
  }

  debug_print(5, ("bp: return %d\n", count < 0 ? 0 : count));
  return count < 0 ? 0 : count;
}

int mutt_count_body_parts (HEADER *hdr, int flags) {
  if (!option (OPTCOUNTATTACH))
    return (0);
  if (hdr->attach_valid && !(flags & M_PARTS_RECOUNT))
    return hdr->attach_total;

  if (AttachAllow || AttachExclude || InlineAllow || InlineExclude)
    hdr->attach_total = count_body_parts(hdr->content, flags | M_PARTS_TOPLEVEL);
  else
    hdr->attach_total = 0;

  hdr->attach_valid = 1;
  return hdr->attach_total;
}
